from __future__ import annotations

from collections.abc import Generator
from pathlib import Path

import cstag
import mappy


def revcomp(sequence: str) -> str:
    complement = {"A": "T", "C": "G", "G": "C", "T": "A"}
    return "".join(complement[nt] for nt in sequence[::-1])


def to_sam(
    path_reference_fasta: str | Path,
    path_query_fastx: str | Path,
    preset: str = "map-ont",
    threads: int = 1,
    cslong: bool = True,
) -> Generator[str]:
    """Align seqences using mappy and Convert PAF to SAM

    Args:
        path_reference_fasta (str): Path of reference fasta
        path_query_fastx (str): Path of query fasta/fastq
        cslong (bool, optional): long formatted CS tag if True. Defaults to True

    Returns:
        list: List of SAM
    """
    path_reference_fasta = str(path_reference_fasta)
    path_query_fastx = str(path_query_fastx)
    # SQ header
    SAM = [f"@SQ\tSN:{n}\tLN:{len(s)}" for n, s, _ in mappy.fastx_read(path_reference_fasta)]
    # Mappy
    ref = mappy.Aligner(path_reference_fasta, preset=preset, n_threads=threads)
    if not ref:
        raise AttributeError(f"Failed to load {path_reference_fasta}")
    for MAPPY_NAME, MAPPY_SEQ, MAPPY_QUAL in mappy.fastx_read(path_query_fastx):
        for hit in ref.map(MAPPY_SEQ, cs=True):
            query_seq = MAPPY_SEQ
            query_qual = MAPPY_QUAL
            # flag
            if hit.is_primary:
                flag = 0 if hit.strand == 1 else 16
            else:
                flag = 2048 if hit.strand == 1 else 2064
            # Append softclips to CIGAR
            cigar = hit.cigar_str
            if hit.q_st > 0:
                softclip = str(hit.q_st) + "S"
                cigar = softclip + cigar if hit.strand == 1 else cigar + softclip
            if len(MAPPY_SEQ) - hit.q_en > 0:
                softclip = str(len(MAPPY_SEQ) - hit.q_en) + "S"
                cigar = cigar + softclip if hit.strand == 1 else softclip + cigar
            # Revcomp
            if hit.strand == -1:
                query_seq = revcomp(query_seq)
                if query_qual:
                    query_qual = query_qual[::-1]
            query_seq = query_seq.upper()
            # cslong
            cs = "cs:Z:" + hit.cs
            if cslong:
                cs = cstag.lengthen(hit.cs, cigar, query_seq)
            # summarize
            alignment = [
                MAPPY_NAME,
                flag,
                hit.ctg,
                hit.r_st + 1,
                hit.mapq,
                cigar,
                "*",
                0,
                0,
                query_seq,
                query_qual,
                cs,
            ]
            alignment = [str(a) for a in alignment]
            SAM.append("\t".join(alignment))
    for record in SAM:
        yield record


def output_sam(
    tmpdir: Path,
    path_fasta: str | Path,
    name_fasta: str,
    path_fastq: str | Path,
    name_fastq: str,
    preset: str = "map-ont",
    threads: int = 1,
):
    sam = to_sam(path_fasta, path_fastq, preset=preset, threads=threads)
    output_sam = Path(tmpdir, "sam", f"{name_fastq}_{preset}_{name_fasta}.sam")
    output_sam.write_text("\n".join(sam))


########################################################################
# Create faidx
########################################################################


def make_faidx(path_fasta: Path | str) -> str:
    fasta = Path(path_fasta).read_text().split()
    name = fasta[0].strip(">")
    length = len("".join(fasta[1:]))
    offset = len(fasta[0]) + 1
    linebase = len(fasta[1])
    linewidth = len(fasta[1]) + 1
    faidx = list(map(str, [name, length, offset, linebase, linewidth]))
    faidx = "\t".join(faidx) + "\n"
    return faidx



from __future__ import annotations

import sys

from DAJIN2.core import core_execute
from DAJIN2.postprocess import report
from DAJIN2.preprocess.validate_inputs import (validate_files,
                                               validate_genome_and_fetch_urls)


def single_execute(arguments: dict[str]):
    ################################################################################
    # Validate contents in the batch file
    ################################################################################
    validate_files(args["sample"], args["control"], args["allele"])
    if "genome" in args:
        UCSC_URL, GOLDENPATH_URL = validate_genome_and_fetch_urls(args["genome"])
    else:
        UCSC_URL, GOLDENPATH_URL = None, None

    print(f"{arguments['control']} is now processing...", file=sys.stderr)
    core_execute.execute_control(arguments)
    print(f"{arguments['sample']} is now processing...", file=sys.stderr)
    core_execute.execute_sample(arguments)
    name = arguments["name"]
    report.report(name)
    print(f"Finished! Open DAJINResults/{name} to see the report.", file=sys.stderr)



from __future__ import annotations

import re
from collections import Counter, defaultdict
from difflib import get_close_matches
from itertools import permutations
from pathlib import Path

import midsv
from scipy import stats
from scipy.spatial.distance import cosine

from DAJIN2.core.preprocess import mappy_align

###############################################################################
# functions
###############################################################################


def extract_knockin_loci(TEMPDIR) -> defaultdict(set):
    """
    Returns:
        defaultdict(set): loci of knockin in each fasta pairs
    """
    fasta_alleles = list(Path(TEMPDIR, "fasta").iterdir())
    fasta_alleles = [f for f in fasta_alleles if f.suffix != ".fai"]
    knockin_alleles = defaultdict(set)
    for pair in list(permutations(fasta_alleles, 2)):
        ref, query = pair
        ref_allele = ref.stem
        alignments = mappy_align.to_sam(ref, query, preset="splice")
        alignments = [a.split("\t") for a in alignments]
        alignments_midsv = midsv.transform(alignments, midsv=False, cssplit=True, qscore=False)[0]
        cssplits = alignments_midsv["CSSPLIT"].split(",")
        knockin_loci = set()
        for i, cs in enumerate(cssplits):
            if cs == "N" or cs.startswith("-"):
                knockin_loci.add(i)
        knockin_alleles[ref_allele] = knockin_loci
    return knockin_alleles


def get_5mer_of_knockin_loci(sequence: str, knockin_loci: list) -> dict:
    sequence_kmer = dict()
    for i in knockin_loci:
        sequence_kmer.update({i + 2: sequence[i : i + 5]})
    return sequence_kmer


def get_5mer_of_sequence(sequence: str) -> dict:
    if len(sequence) <= 5:
        return [sequence]
    sequence_kmer = dict()
    for i in range(len(sequence) - 5):
        sequence_kmer.update({i + 2: sequence[i : i + 5]})
    return sequence_kmer


def get_idx_of_similar_5mers(knockin_kmer: dict, sequence_kmer: dict, knockin_loci: set, n=100) -> defaultdict(set):
    idx_of_similar_5mers = defaultdict(list)
    for locus, kmer in knockin_kmer.items():
        idxes = set()
        for similar_kmer in get_close_matches(kmer, sequence_kmer.values(), n=n, cutoff=0.0):
            for idx in (key for key, val in sequence_kmer.items() if val == similar_kmer):
                if set(range(idx - 2, idx + 3)) & knockin_loci:
                    continue
                idxes.add(idx)
        idx_of_similar_5mers[locus] = idxes
    return idx_of_similar_5mers


def count_indel_5mer(cssplits_transposed: list, indexes: list(int)) -> defaultdict(dict):
    count_5mer = defaultdict(dict)
    for i in indexes:
        cssplits_5mer = cssplits_transposed[i - 2 : i + 3]
        count = {"ins": [1] * 5, "del": [1] * 5, "sub": [1] * 5}
        for j, cs in enumerate(cssplits_5mer):
            counter = Counter(cs)
            for key, cnt in counter.items():
                if key.startswith("=") or key == "N" or re.search(r"a|c|g|t|n", key):
                    continue
                if key.startswith("+"):
                    count["ins"][j] += cnt
                elif key.startswith("-"):
                    count["del"][j] += cnt
                elif key.startswith("*"):
                    count["sub"][j] += cnt
        count_5mer[i] = count
    return count_5mer


def replace_errors_to_match(cssplits_sample: list, sequence_errors: defaultdict(set), sequence: str):
    cssplits_replaced = []
    for cssplits in cssplits_sample:
        cssplits_copy = cssplits.copy()
        for i, error in sequence_errors.items():
            cssplits_5mer = cssplits_copy[i - 2 : i + 3]
            for j, mer in enumerate(cssplits_5mer):
                match_seq = "=" + sequence[i - 2 + j]
                if "ins" in error and mer.startswith("+"):
                    cssplits_5mer[j] = match_seq
                if "del" in error and mer.startswith("-"):
                    cssplits_5mer[j] = match_seq
                if "del" in error and mer.startswith("*"):
                    cssplits_5mer[j] = match_seq
            cssplits_copy[i - 2 : i + 3] = cssplits_5mer
        cssplits_replaced.append(cssplits_copy)
    return cssplits_replaced


##########################################################
# main
##########################################################


def execute(TEMPDIR, FASTA_ALLELES, CONTROL_NAME, SAMPLE_NAME):
    knockin_alleles = extract_knockin_loci(TEMPDIR)
    for allele, sequence in FASTA_ALLELES.items():
        sequence = FASTA_ALLELES[allele]
        knockin_loci = knockin_alleles[allele]
        midsv_sample = midsv.read_jsonl(Path(TEMPDIR, "midsv", f"{SAMPLE_NAME}_{allele}.jsonl"))
        midsv_control = midsv.read_jsonl(Path(TEMPDIR, "midsv", f"{CONTROL_NAME}_{allele}.jsonl"))
        cssplits_sample = [m["CSSPLIT"].split(",") for m in midsv_sample]
        cssplits_control = [m["CSSPLIT"].split(",") for m in midsv_control]
        # Split the knock-in sequence and control into 5 mer and find 100 similar sequences in the control
        knockin_kmer = get_5mer_of_knockin_loci(sequence, knockin_loci)
        sequence_kmer = get_5mer_of_sequence(sequence)
        idx_of_similar_5mers = get_idx_of_similar_5mers(knockin_kmer, sequence_kmer, knockin_loci, n=100)
        # Find the number of indels in 5mer of similar sequence
        count_5mer_similar_sequences = defaultdict(dict)
        cssplits_transposed = [list(t) for t in zip(*cssplits_control)]
        for i, indexes in idx_of_similar_5mers.items():
            count_5mer_similar_sequences[i] = count_indel_5mer(cssplits_transposed, indexes)
        # Find the number of indels in 5mer of Knock-in sequence
        cssplits_transposed = [list(t) for t in zip(*cssplits_sample)]
        count_5mer_knockin = count_indel_5mer(cssplits_transposed, knockin_loci)
        # If there is an error profile similar to the Knock-in sequence and similar sequences, consider it a sequencing error
        coverage_sample = len(midsv_sample)
        coverage_control = len(midsv_control)
        sequence_errors = defaultdict(set)
        for i, count_knockin in count_5mer_knockin.items():
            knockin_mutation = defaultdict(list)
            for mutation in ["ins", "del", "sub"]:
                knockin_mutation[mutation] = [c / coverage_sample for c in count_knockin[mutation]]
            count_control = count_5mer_similar_sequences[i]
            for _, count in count_control.items():
                for mutation in ["ins", "del", "sub"]:
                    knockin = knockin_mutation[mutation]
                    control = [c / coverage_control for c in count[mutation]]
                    distance = 1 - cosine(knockin, control)
                    _, pvalue = stats.ttest_ind(knockin, control, equal_var=False)
                    if distance > 0.7 and pvalue > 0.01:
                        sequence_errors[i].add(mutation)
        # Correct sequencing errors in knock-in sequences
        cssplits_replaced = replace_errors_to_match(cssplits_sample, sequence_errors, sequence)
        # Replace CSSPLIT
        cssplits_corrected = [",".join(cs) for cs in cssplits_replaced]
        for i, cssplits in enumerate(cssplits_corrected):
            midsv_sample[i]["CSSPLIT"] = cssplits
        # Save to jsonl
        midsv.write_jsonl(midsv_sample, Path(TEMPDIR, "midsv", f"{SAMPLE_NAME}_{allele}.jsonl"))



from __future__ import annotations

import re
import json
import random
import tempfile
import subprocess
import numpy as np
from pathlib import Path
from itertools import islice
from typing import Generator
from collections import Counter, defaultdict


def _generate_3mer(midsv_sample: Generator[dict[str, str]]) -> Generator[str]:
    for cssplits in (cs["CSSPLIT"].split(",") for cs in midsv_sample):
        cssplits_kmer = []
        for idx_seq, cs in enumerate(cssplits):
            if idx_seq == 0:
                two_prior = cs
                cssplits_kmer.append(cs)
                continue
            if idx_seq == 1:
                one_prior = cs
                cssplits_kmer.append(cs)
                continue
            cssplits_kmer.append(two_prior + "," + one_prior + "," + cs)
            two_prior, one_prior = one_prior, cs
        yield cssplits_kmer


def _get_path_of_transposed_matrix(TEMPDIR, cssplits_sample: str, chunksize: int = 100, sep: str = "\t") -> str:
    # Write cssplits
    with tempfile.NamedTemporaryFile(dir=TEMPDIR) as fp:
        temp_cssplits = fp.name
    with open(temp_cssplits, "w") as out:
        for cssplits in cssplits_sample:
            out.write(sep.join(cssplits) + "\n")
    # Transpose by subprocess
    with tempfile.NamedTemporaryFile(dir=TEMPDIR) as fp:
        temp_output = fp.name
    with tempfile.NamedTemporaryFile(dir=TEMPDIR) as fp:
        temp_transposed_chunk = fp.name
    with tempfile.NamedTemporaryFile(dir=TEMPDIR) as fp:
        temp_pasted = fp.name
    with tempfile.NamedTemporaryFile(dir=TEMPDIR) as fp:
        output_path = fp.name
    with open(temp_cssplits, "r") as f:
        while True:
            next_n_lines = [line.strip().split(sep) for line in islice(f, chunksize)]
            if not next_n_lines:
                break
            with open(temp_transposed_chunk, "w") as out:
                for line in zip(*next_n_lines):
                    out.write(sep.join(line) + "\n")
            if not Path(temp_output).exists():
                subprocess.call(f"mv {temp_transposed_chunk} {temp_output}", shell=True)
            else:
                subprocess.call(f"paste -d '{sep}' {temp_output} {temp_transposed_chunk} > {temp_pasted}", shell=True)
                subprocess.call(f"cat {temp_pasted} > {temp_output}", shell=True)
        subprocess.call(f"cat {temp_output} > {output_path}", shell=True)
    return output_path


def _read_cssplits_as_tsv(input_path: str, sep: str = "\t"):
    with open(input_path, "r") as f:
        for line in f:
            yield line.strip().split(sep)


def _sampling_cssplits(transposed_midsv, mutation_loci) -> Generator:
    for idx_seq, cssplits_3mers in enumerate(transposed_midsv):
        sampling_2mer = defaultdict(list)
        if idx_seq == 0 or idx_seq == 1:
            yield defaultdict(Counter)
            continue
        for cssplits_3mer in cssplits_3mers:
            two_prior, one_prior, cs = cssplits_3mer.split(",")
            if cs[0] in mutation_loci[idx_seq] or cs.startswith("=") or cs == "N" or re.search(r"a|c|g|t|n", cs):
                previous_2mer = two_prior + "," + one_prior
                sampling_2mer[previous_2mer].append(cs)
            one_prior, two_prior = two_prior, cs
        yield {mer: Counter(samp) for mer, samp in sampling_2mer.items()}


def _correct_errors_cssplits(transposed_midsv, mutation_loci, sampling_cssplits):
    random.seed(1)
    for idx_seq, (cssplits_3mers, sampling_cssplit) in enumerate(zip(transposed_midsv, sampling_cssplits)):
        if idx_seq == 0 or idx_seq == 1:
            yield [cs for cs in cssplits_3mers]
            continue
        cssplits_corrected = []
        for cssplits_3mer in cssplits_3mers:
            two_prior, one_prior, cs = cssplits_3mer.split(",")
            if not (cs[0] in mutation_loci[idx_seq] or cs.startswith("=") or cs == "N" or re.search(r"a|c|g|t|n", cs)):
                previous_2mer = f"{two_prior},{one_prior}"
                try:
                    sampling = sampling_cssplit[previous_2mer]
                except KeyError:
                    cs = "N"
                else:
                    cs = random.choices(*zip(*sampling.items()))[0]
            cssplits_corrected.append(cs)
        yield cssplits_corrected


def _combine_with_qname_cssplits(midsv_sample, cssplits_corrected):
    for samp, cssplits in zip(midsv_sample, cssplits_corrected):
        yield {"QNAME": samp["QNAME"], "CSSPLIT": ",".join(cssplits)}


def read_json(filepath) -> Generator[dict[str, str]]:
    with open(filepath, "r") as f:
        for line in f:
            yield json.loads(line)


###############################################################################
# main
###############################################################################


def correct_sequence_error(TEMPDIR, FASTA_ALLELES, SAMPLE_NAME, CONTROL_NAME, MUTATION_LOCI_ALLELES) -> None:
    for allele in FASTA_ALLELES:
        mutation_loci = MUTATION_LOCI_ALLELES[allele]
        for name in [SAMPLE_NAME, CONTROL_NAME]:
            filepath = Path(TEMPDIR, "midsv", f"{name}_{allele}.json")
            midsv_3mer = _generate_3mer(read_json(filepath))
            path_transposed = _get_path_of_transposed_matrix(TEMPDIR, midsv_3mer, chunksize=1000)
            sampling = _sampling_cssplits(_read_cssplits_as_tsv(path_transposed), mutation_loci)
            cssplits_corrected_ = _correct_errors_cssplits(
                _read_cssplits_as_tsv(path_transposed), mutation_loci, sampling
            )
            path_transposed_corrected = _get_path_of_transposed_matrix(TEMPDIR, cssplits_corrected_, chunksize=1000)
            qname_cssplits = _combine_with_qname_cssplits(
                read_json(filepath), _read_cssplits_as_tsv(path_transposed_corrected)
            )
            out_filepath = Path(TEMPDIR, "midsv_corrected", f"{name}_{allele}.json")
            with open(out_filepath, "wt", encoding="utf-8") as f:
                for data in qname_cssplits:
                    f.write(json.dumps(data) + "\n")


# from __future__ import annotations

# import random
# import re
# from collections import Counter, defaultdict
# from typing import Generator
# import json
# from pathlib import Path


# def _sampling_cssplits(midsv_sample: dict[list[str, str]], mutation_loci: dict[str, set[int]]) -> list[dict[Counter]]:
#     sampling = [defaultdict(list) for _ in range(len(mutation_loci))]
#     for cssplits in (cs["CSSPLIT"] for cs in midsv_sample):
#         for idx_seq, cs in enumerate(cssplits.split(",")):
#             if idx_seq == len(mutation_loci) - 1:
#                 break
#             if idx_seq == 0:
#                 one_prior = cs
#                 continue
#             if idx_seq == 1:
#                 two_prior = cs
#                 continue
#             if cs[0] in mutation_loci[idx_seq] or cs.startswith("=") or cs == "N" or re.search(r"a|c|g|t|n", cs):
#                 prev_cs = one_prior + "," + two_prior
#                 sampling[idx_seq][prev_cs].append(cs)
#             one_prior, two_prior = two_prior, cs
#     sampling_cssplits = []
#     for samp in sampling:
#         samp_counter = dict()
#         for key, val in samp.items():
#             val = Counter(val)
#             samp_counter.update({key: val})
#         sampling_cssplits.append(samp_counter)
#     return sampling_cssplits


# def _correct_errors(
#     midsv_sample: Generator, mutation_loci: list(set(str)), sampling_cssplits: list(defaultdict(Counter))
# ) -> Generator[dict[str, str]]:
#     random.seed(1)
#     for samp in midsv_sample:
#         qname = samp["QNAME"]
#         cssplits = samp["CSSPLIT"].split(",")
#         for idx_seq, cs in enumerate(cssplits):
#             if idx_seq == len(mutation_loci) - 1:
#                 break
#             if idx_seq == 0:
#                 two_prior = cs
#                 continue
#             if idx_seq == 1:
#                 one_prior = cs
#                 continue
#            if not (cs[0] in mutation_loci[idx_seq] or cs.startswith("=") or cs == "N" or re.search(r"a|c|g|t|n", cs)):
#                 previous_2mer = f"{two_prior},{one_prior}"
#                 sampling = sampling_cssplits[idx_seq][previous_2mer]
#                 if sampling == Counter():
#                     cssplits[idx_seq] = "N"
#                     continue
#                 cssplits[idx_seq] = random.choices(*zip(*sampling.items()))[0]
#             two_prior, one_prior = one_prior, cs
#         cssplits_joined = ",".join(cssplits)
#         yield {"QNAME": qname, "CSSPLIT": cssplits_joined}


# def read_json(filepath) -> Generator[dict[str, str]]:
#     with open(filepath, "r") as f:
#         for line in f:
#             yield json.loads(line)


# def write_midsv(filepath, data) -> None:
#     with open(filepath, "wt", encoding="utf-8") as f:
#         for d in data:
#             f.write(json.dumps(d) + "\n")


# ###############################################################################
# # main
# ###############################################################################


# def correct_sequence_error(TEMPDIR, FASTA_ALLELES, SAMPLE_NAME, CONTROL_NAME, MUTATION_LOCI_ALLELES) -> None:
#     for allele in FASTA_ALLELES:
#         filepath_sample = Path(TEMPDIR, "midsv", f"{SAMPLE_NAME}_{allele}.json")
#         filepath_control = Path(TEMPDIR, "midsv", f"{CONTROL_NAME}_{allele}.json")
#         # Extract mutation loci
#         mutation_loci = MUTATION_LOCI_ALLELES[allele]
#         # Correct sequence errors
#         sampling_sample = _sampling_cssplits(read_json(filepath_sample), mutation_loci)
#         sampling_control = _sampling_cssplits(read_json(filepath_control), mutation_loci)
#         midsv_corected_sample = _correct_errors(read_json(filepath_sample), mutation_loci, sampling_sample)
#         midsv_corected_control = _correct_errors(read_json(filepath_control), mutation_loci, sampling_control)
#         # Output corrected midsv
#         filepath_sample = Path(TEMPDIR, "midsv_corrected", f"{SAMPLE_NAME}_{allele}.json")
#         filepath_control = Path(TEMPDIR, "midsv_corrected", f"{CONTROL_NAME}_{allele}.json")
#         write_midsv(filepath_sample, midsv_corected_sample)
#         write_midsv(filepath_control, midsv_corected_control)



from __future__ import annotations

import json
import re
from itertools import chain, groupby
from pathlib import Path
from typing import Generator

import midsv


def _load_sam(path_sam: str | Path) -> Generator[list[str]]:
    return midsv.read_sam(path_sam)


def _split_cigar(CIGAR: str) -> list[str]:
    cigar = re.split(r"([MIDNSH=X])", CIGAR)
    n = len(cigar)
    cigar_split = []
    for i, j in zip(range(0, n, 2), range(1, n, 2)):
        cigar_split.append(cigar[i] + cigar[j])
    return cigar_split


def _call_alignment_length(CIGAR: str) -> int:
    cigar_split = _split_cigar(CIGAR)
    alignment_length = 0
    for c in cigar_split:
        if re.search(r"[MDN=X]", c[-1]):
            alignment_length += int(c[:-1])
    return alignment_length


def _has_inversion_in_splice(CIGAR: str) -> bool:
    is_splice = False
    is_insertion = False
    for cigar in _split_cigar(CIGAR):
        if cigar.endswith("I"):
            is_insertion = True
            continue
        if is_insertion and cigar.endswith("N"):
            is_splice = True
            break
        else:
            is_insertion = False
    return is_splice


def _extract_qname_of_map_ont(sam_ont: Generator[list[str]], sam_splice: Generator[list[str]]) -> set():
    """Extract qname of reads from `map-ont` when:
    - no inversion signal in `splice` alignment (insertion + deletion)
    - single read
    - long alignment length
    """
    dict_alignments_splice = {s[0]: s for s in sam_splice if not s[0].startswith("@")}
    alignments_ont = [s for s in sam_ont if not s[0].startswith("@")]
    alignments_ont.sort(key=lambda x: x[0])
    qname_of_map_ont = set()
    for qname_ont, group in groupby(alignments_ont, key=lambda x: x[0]):
        alignment_ont = list(group)
        if qname_ont not in dict_alignments_splice:
            qname_of_map_ont.add(qname_ont)
            continue
        alignment_splice = dict_alignments_splice[qname_ont]
        if _has_inversion_in_splice(alignment_splice[5]):
            qname_of_map_ont.add(qname_ont)
            continue
        if len(alignment_ont) != 1:
            continue
        alignment_ont = alignment_ont[0]
        alignment_length_ont = _call_alignment_length(alignment_ont[5])
        alignment_length_splice = _call_alignment_length(alignment_splice[5])
        if alignment_length_ont >= alignment_length_splice:
            qname_of_map_ont.add(qname_ont)
    return qname_of_map_ont


def _extract_sam(sam: Generator[list[str]], qname_of_map_ont: set, preset: str = "map-ont") -> Generator[list[str]]:
    for alignment in sam:
        if alignment[0].startswith("@"):
            yield alignment
        if preset == "map-ont":
            if alignment[0] in qname_of_map_ont:
                yield alignment
        else:
            if alignment[0] not in qname_of_map_ont:
                yield alignment


def _midsv_transform(sam: Generator[list[str]]) -> Generator[list[str]]:
    for midsv_sample in midsv.transform(sam, midsv=False, cssplit=True, qscore=False):
        yield midsv_sample


def _replaceNtoD(midsv_sample: Generator[list[str]], sequence: str) -> Generator[dict[str, str]]:
    for samp in midsv_sample:
        qname = samp["QNAME"]
        cssplits = samp["CSSPLIT"].split(",")
        # extract right/left index of the end of sequential Ns
        left_idx_n = 0
        for cs in cssplits:
            if cs != "N":
                break
            left_idx_n += 1
        right_idx_n = 0
        for cs in cssplits[::-1]:
            if cs != "N":
                break
            right_idx_n += 1
        right_idx_n = len(cssplits) - right_idx_n - 1
        # replace sequential Ns within the sequence
        for j, (cs, seq) in enumerate(zip(cssplits, sequence)):
            if left_idx_n <= j <= right_idx_n and cs == "N":
                cssplits[j] = f"-{seq}"
        yield {"QNAME": qname, "CSSPLIT": ",".join(cssplits)}


def call_midsv(TEMPDIR: Path | str, FASTA_ALLELES: dict, SAMPLE_NAME: str) -> None:
    for allele, sequence in FASTA_ALLELES.items():
        path_ont = Path(TEMPDIR, "sam", f"{SAMPLE_NAME}_map-ont_{allele}.sam")
        path_splice = Path(TEMPDIR, "sam", f"{SAMPLE_NAME}_splice_{allele}.sam")
        qname_of_map_ont = _extract_qname_of_map_ont(_load_sam(path_ont), _load_sam(path_splice))
        sam_of_map_ont = _extract_sam(_load_sam(path_ont), qname_of_map_ont, preset="map-ont")
        sam_of_splice = _extract_sam(_load_sam(path_splice), qname_of_map_ont, preset="splice")
        sam_chained = chain(sam_of_map_ont, sam_of_splice)
        midsv_chaind = _midsv_transform(sam_chained)
        midsv_sample = _replaceNtoD(midsv_chaind, sequence)
        filepath = Path(TEMPDIR, "midsv", f"{SAMPLE_NAME}_{allele}.json")
        with open(filepath, "wt", encoding="utf-8") as f:
            for data in midsv_sample:
                f.write(json.dumps(data) + "\n")



from __future__ import annotations

import os
import re
from pathlib import Path
from urllib.request import urlopen

import mappy
import midsv
import wslPath

########################################################################
# Make directories
########################################################################


def make_directories(TEMPDIR: Path, SUBDIRS: list[str], SUBDIRS_REPORT: list[str], SAMPLE_NAME: str, CONTROL_NAME: str):
    for subdir in SUBDIRS:
        Path(TEMPDIR, subdir).mkdir(parents=True, exist_ok=True)
    for reportdir in SUBDIRS_REPORT:
        Path(TEMPDIR, "report", reportdir, SAMPLE_NAME).mkdir(parents=True, exist_ok=True)
    Path(TEMPDIR, "report", "BAM", CONTROL_NAME).mkdir(parents=True, exist_ok=True)
    Path(TEMPDIR, "cache", ".igvjs").mkdir(parents=True, exist_ok=True)


########################################################################
# Convert Path
########################################################################


def convert_to_posix_path(path: str) -> str:
    try:
        path = wslPath.toPosix(path)
    except ValueError:
        pass
    return str(path)


########################################################################
# Extract basename
########################################################################


def extract_basename(fastq_path: str) -> str:
    name = Path(fastq_path).name
    name = re.sub(r"\..*$", "", name)
    name = name.lstrip()
    if not name:
        raise AttributeError(f"{fastq_path} is not valid file name")
    else:
        return re.sub(r'[\\|/|:|?|.|,|\'|"|<|>|\| |]', "-", name)


########################################################################
# Convert allele file to dictionary type fasta format
########################################################################


def dictionize_allele(path_fasta: str) -> dict:
    header, sequence = [], []
    for name, seq, _ in mappy.fastx_read(path_fasta):
        name = name.lstrip()
        if not name:
            raise AttributeError(f"{path_fasta} contains an empty header")
        name = re.sub(r'[\\|/|:|?|.|,|\'|"|<|>|\| |]', "-", name)
        header.append(name)
        sequence.append(seq.upper())
    return {h: s for h, s in zip(header, sequence)}


########################################################################
# Update threads
########################################################################


def update_threads(threads):
    threads_updated = min(int(threads), os.cpu_count() - 1)
    threads_updated = max(1, threads_updated)
    return threads_updated


########################################################################
# Fetch genome coodinate and chrom size
########################################################################


# TODO TEST
def fetch_coodinate(genome: str, ucsc_url: str, seq: str) -> dict:
    ucsc_blat = f"{ucsc_url}/cgi-bin/hgBlat?db={genome}&type=BLAT&userSeq={seq}"
    request = urlopen(ucsc_blat).read().decode("utf8").split("\n")
    coodinate = [x for x in request if x.count("100.0%")]
    if not coodinate:
        raise AttributeError(f"{seq} is not found in {genome}")
    else:
        coodinate = coodinate[0].split()
        return {"chr": coodinate[-5], "start": int(coodinate[-3]), "end": int(coodinate[-2]), "strand": coodinate[-4]}


def fetch_chrom_size(chrom: str, genome: str, goldenpath_url: str) -> int:
    url = f"{goldenpath_url}/{genome}/bigZips/{genome}.chrom.sizes"
    request = urlopen(url).read().decode("utf8").split("\n")
    for req in request:
        req = req.split("\t")
        if chrom == req[0]:
            chrom_size = int(req[1])
    return chrom_size


def cache_coodinates_and_chromsize(TEMPDIR, GENOME, GENOME_COODINATES, CHROME_SIZE):
    """
    Save (1) genome_symbol.txt, (2) genome_coodinates.jsonl, (3) chrome_size.txt
    """
    # Save info to the cache directory
    Path(TEMPDIR, "cache", "genome_symbol.txt").write_text(GENOME + "\n")
    midsv.write_jsonl([GENOME_COODINATES], Path(TEMPDIR, "cache", "genome_coodinates.jsonl"))
    Path(TEMPDIR, "cache", "chrome_size.txt").write_text(str(CHROME_SIZE))
    # Save info to the .igvjs directory
    Path(TEMPDIR, "report", ".igvjs", "genome_symbol.txt").write_text(GENOME + "\n")
    midsv.write_jsonl([GENOME_COODINATES], Path(TEMPDIR, "report", ".igvjs", "genome_coodinates.jsonl"))



from __future__ import annotations

import warnings
from collections import Counter
from typing import Generator
from pathlib import Path
import json
from itertools import chain

import numpy as np
from sklearn.decomposition import PCA
from sklearn.exceptions import ConvergenceWarning
from sklearn.mixture import GaussianMixture

from DAJIN2.core.clustering.merge_clusters import merge_clusters

warnings.filterwarnings("ignore", category=RuntimeWarning)
warnings.filterwarnings("ignore", category=ConvergenceWarning)

###############################################################################
# Dimension reduction
###############################################################################


def reduce_dimension(scores_sample: Generator[list], scores_control: Generator[list]) -> np.array:
    scores = list(chain(scores_sample, scores_control))
    pca = PCA(n_components=20).fit(scores)
    return pca.transform(scores)


def optimize_labels(X: np.array, coverage_sample, coverage_control) -> list[int]:
    n_components = min(20, coverage_sample + coverage_control)
    for i in range(1, n_components):
        np.random.seed(seed=1)
        labels = GaussianMixture(n_components=i, random_state=1).fit_predict(X)
        labels = labels.tolist()
        labels_sample = labels[:coverage_sample]
        labels_control = labels[coverage_sample:]
        labels_merged = merge_clusters(labels_control, labels_sample)
        # print(i, Counter(labels_sample), Counter(labels_control), Counter(labels_merged)) # ! DEBUG
        # Reads < 1% in the control are considered clustering errors and are not counted
        count_control = Counter(labels_control)
        num_labels_control = sum(1 for reads in count_control.values() if reads / coverage_control > 0.01)
        # Report the number of clusters in SAMPLE when the number of clusters in CONTROL is split into more than one.
        if num_labels_control > 1:
            return labels_merged
        else:
            labels_results = labels_merged
    return labels_results


def read_json(filepath: Path | str) -> Generator[dict[str, str]]:
    with open(filepath, "r") as f:
        for line in f:
            yield json.loads(line)


def count_newlines(filepath: Path | str) -> int:
    def _make_gen(reader):
        while True:
            b = reader(2**16)
            if not b:
                break
            yield b

    with open(filepath, "rb") as f:
        count = sum(buf.count(b"\n") for buf in _make_gen(f.raw.read))
    return count


###############################################################################
# main
###############################################################################


def return_labels(path_score_sample: Path | str, path_score_control: Path | str) -> list[int]:
    np.random.seed(seed=1)
    # Remove abnormal minor reads from control
    X_control = reduce_dimension([], read_json(path_score_control))
    labels = GaussianMixture(n_components=2, random_state=1).fit_predict(X_control)
    label_most = Counter(labels).most_common()[0][0]
    # Subset control to 1000 reads
    scores_control_subset = []
    count = 0
    for label, score in zip(labels, read_json(path_score_control)):
        if label == label_most:
            scores_control_subset.append(score)
            count += 1
        if count == 1000:
            break
    X = reduce_dimension(read_json(path_score_sample), scores_control_subset)
    coverage_sample = count_newlines(path_score_sample)
    coverage_control = len(scores_control_subset)
    labels = optimize_labels(X, coverage_sample, coverage_control)
    return labels


# def return_labels(scores_sample: Generator[list[float]], scores_control: Generator[list[float]]) -> list[int]:
#     np.random.seed(seed=1)
#     X_control = reduce_dimension([], scores_control)
#     labels = GaussianMixture(n_components=2, random_state=1).fit_predict(X_control)
#     label_most = Counter(labels).most_common()[0][0]
#     scores_control_subset = [s for label, s in zip(labels, scores_control) if label == label_most][:1000]
#     X = reduce_dimension(scores_sample, scores_control_subset)
#     labels = optimize_labels(X, scores_sample, scores_control_subset)
#     return labels



from __future__ import annotations

import re
from collections import defaultdict
import json
import numpy as np
from pathlib import Path
from typing import Generator
from scipy import stats
from scipy.spatial import distance
from sklearn.neighbors import LocalOutlierFactor
from DAJIN2.core.preprocess.extract_errors_in_homopolymer import extract_errors_in_homopolymer


def read_midsv(filepath) -> Generator[dict[str, str]]:
    with open(filepath, "r") as f:
        for line in f:
            yield json.loads(line)


def _make_gen(reader):
    while True:
        b = reader(2**16)
        if not b:
            break
        yield b


def count_newlines(filepath):
    with open(filepath, "rb") as f:
        count = sum(buf.count(b"\n") for buf in _make_gen(f.raw.read))
    return count


def _count_indels(midsv_sample, len_sequence: int) -> dict[str, list[int]]:
    count = {"+": [0] * len_sequence, "-": [0] * len_sequence, "*": [0] * len_sequence}
    for samp in midsv_sample:
        for i, cs in enumerate(samp["CSSPLIT"].split(",")):
            if cs.startswith("=") or cs == "N" or re.search(r"a|c|g|t|n", cs):
                continue
            if cs.startswith("+"):
                # count["+"][i] += len(cs.split("|"))
                count["+"][i] += 1
            elif cs.startswith("-"):
                count["-"][i] += 1
            elif cs.startswith("*"):
                count["*"][i] += 1
    return count


def _normalize_indels(count: dict[str, list[int]], coverage: int) -> dict[str, np.array]:
    count_normalized = dict()
    for mut in count:
        counts = np.array(count[mut])
        counts_norm = (counts / coverage) * 10**6
        count_normalized[mut] = counts_norm.astype("uint32")
    return count_normalized


def _split_kmer(indels: dict[str, np.array], kmer: int = 10) -> dict[str, np.array]:
    results = defaultdict(list)
    center = kmer // 2
    for mut, value in indels.items():
        for i in range(len(value)):
            if center <= i <= len(value) - center:
                start = i - center
                if kmer % 2 == 0:
                    end = i + center
                else:
                    end = i + center + 1
                results[mut].append(value[start:end])
            else:
                results[mut].append(np.array([0] * kmer, dtype="uint32"))
    return results


def _extract_anomaly_loci(indels_kmer_sample: dict, indels_kmer_control: dict) -> dict[str, set[int]]:
    anomaly_loci = dict()
    clf = LocalOutlierFactor(novelty=True, n_neighbors=5)
    for mut in indels_kmer_sample.keys():
        loci = set()
        values_control = indels_kmer_control[mut]
        values_sample = indels_kmer_sample[mut]
        index = -1
        for i, (value_control, value_sample) in enumerate(zip(values_control, values_sample)):
            if i == index:
                continue
            clf.fit(value_control.reshape(-1, 1))
            pred = clf.predict(value_sample.reshape(-1, 1))
            if pred[5] == -1:
                loci.add(i)
            # If the next base is not -1, do not validate the next base because the next base is not an outlier.
            if pred[6] == 1:
                index = i + 1
        anomaly_loci.update({mut: loci})
    return anomaly_loci


def _extract_dissimilar_loci(indels_kmer_sample: dict, indels_kmer_control: dict) -> dict[str, set]:
    """
    Comparing Sample and Control, the 1. 'high similarity',  2. 'similar mean' and
    3. 'similar variance' are considered as sequence errors.
    """
    results = dict()
    for mut in indels_kmer_sample:
        values_sample = indels_kmer_sample[mut]
        values_control = indels_kmer_control[mut]
        # Calculate cosine similarity: 1 means exactly same, 0 means completely different.
        # When calculating cossim, uint32 returns inaccurate results so convert to float64
        cossim = [
            1 - distance.cosine(x.astype("float64"), y.astype("float64")) for x, y in zip(values_sample, values_control)
        ]
        # Perform T-test: nan means exactly same, p > 0.05 means similar in average.
        t_pvalues = [stats.ttest_ind(x, y, equal_var=False)[1] for x, y in zip(values_sample, values_control)]
        t_pvalues = [1 if np.isnan(t) else t for t in t_pvalues]
        # Perform F-test: p > 0.05 means similar in variance.
        f_pvalues = [stats.bartlett(x, y)[1] for x, y in zip(values_sample, values_control)]
        # if pvalue == nan or pval > 0.05, samples and controls are similar.
        dissimilar_loci = set()
        for i, (sim, t_pval, f_pval) in enumerate(zip(cossim, t_pvalues, f_pvalues)):
            flag_seqerror = False
            if sim > 0.90 and t_pval > 0.05 and f_pval > 0.05:
                flag_seqerror = True
            if flag_seqerror is False:
                dissimilar_loci.add(i)
        results.update({mut: dissimilar_loci})
    return results


def _transpose_mutation_loci(mutation_loci, len_sequence):
    mutation_loci_transposed = [set() for _ in range(len_sequence)]
    for mut, idx_mutation in mutation_loci.items():
        for i, loci in enumerate(mutation_loci_transposed):
            if i in idx_mutation:
                loci.add(mut)
    return mutation_loci_transposed


###########################################################
# main
###########################################################


def extract_mutation_loci(
    TEMPDIR: Path, FASTA_ALLELES: dict, SAMPLE_NAME: str, CONTROL_NAME: str
) -> dict[str, list[set[str]]]:
    MUTATION_LOCI_ALLELES = dict()
    for allele, sequence in FASTA_ALLELES.items():
        filepath_sample = Path(TEMPDIR, "midsv", f"{SAMPLE_NAME}_{allele}.json")
        filepath_control = Path(TEMPDIR, "midsv", f"{CONTROL_NAME}_{allele}.json")
        coverage_sample = count_newlines(filepath_sample)
        coverage_control = count_newlines(filepath_control)
        indels_sample = _count_indels(read_midsv(filepath_sample), len(sequence))
        indels_control = _count_indels(read_midsv(filepath_control), len(sequence))
        indels_sample_normalized = _normalize_indels(indels_sample, coverage_sample)
        indels_control_normalized = _normalize_indels(indels_control, coverage_control)
        indels_kmer_sample = _split_kmer(indels_sample_normalized, kmer=10)
        indels_kmer_control = _split_kmer(indels_control_normalized, kmer=10)
        anomaly_loci = _extract_anomaly_loci(indels_kmer_sample, indels_kmer_control)
        dissimilar_loci = _extract_dissimilar_loci(indels_kmer_sample, indels_kmer_control)
        errors_in_homopolymer = extract_errors_in_homopolymer(indels_sample, indels_control, sequence)
        mutation_loci = dict()
        for mut in anomaly_loci:
            mutation_loci.update({mut: (anomaly_loci[mut] & dissimilar_loci[mut]) - errors_in_homopolymer})
        mutation_loci_transposed = _transpose_mutation_loci(mutation_loci, len(sequence))
        MUTATION_LOCI_ALLELES.update({allele: mutation_loci_transposed})
    return MUTATION_LOCI_ALLELES



from __future__ import annotations

from itertools import permutations
from pathlib import Path

import midsv

from DAJIN2.core.preprocess import mappy_align


def extract_knockin_loci(TEMPDIR: str | Path) -> dict(set(int)):
    """
    Returns:
        dict(set): loci of knockin in each fasta pairs
    """
    fasta_alleles = list(Path(TEMPDIR, "fasta").iterdir())
    fasta_alleles = [f for f in fasta_alleles if f.suffix != ".fai"]
    KNOCKIN_LOCI_ALLELES = dict()
    for pair in permutations(fasta_alleles, 2):
        ref, query = pair
        ref_allele = ref.stem
        alignments = mappy_align.to_sam(ref, query, preset="splice")
        alignments = [a.split("\t") for a in alignments]
        alignments_midsv = midsv.transform(alignments, midsv=False, cssplit=True, qscore=False)[0]
        cssplits = alignments_midsv["CSSPLIT"].split(",")
        knockin_loci = set()
        for i, cs in enumerate(cssplits):
            if cs == "N" or cs.startswith("-"):
                knockin_loci.add(i)
        KNOCKIN_LOCI_ALLELES.update({ref_allele: knockin_loci})
    return KNOCKIN_LOCI_ALLELES



from DAJIN2.core.preprocess import (
    correct_knockin,
    format_inputs,
    mappy_align,
    validate_inputs,
)
from DAJIN2.core.preprocess.call_midsv import call_midsv
from DAJIN2.core.preprocess.correct_sequence_error import correct_sequence_error
from DAJIN2.core.preprocess.extract_knockin_loci import extract_knockin_loci
from DAJIN2.core.preprocess.extract_mutation_loci import extract_mutation_loci
from DAJIN2.core.preprocess.replace_NtoD import replace_NtoD



from __future__ import annotations

import os
import sys
import json
import pickle
import resource
import shutil
from datetime import datetime
from pathlib import Path

import midsv

from DAJIN2.core import classification, clustering, consensus, preprocess, report

# limit max memory usage
mem_bytes = os.sysconf("SC_PAGE_SIZE") * os.sysconf("SC_PHYS_PAGES")
resource.setrlimit(resource.RLIMIT_DATA, (int(mem_bytes * 9 / 10), -1))


def _parse_arguments(arguments: dict):
    SAMPLE: str = arguments["sample"]
    CONTROL: str = arguments["control"]
    ALLELE: str = arguments["allele"]
    NAME: str = arguments["name"]
    THREADS: int = arguments["threads"]
    if "genome" in arguments:
        GENOME: str = arguments["genome"]
        URL_UCSC = arguments["ucsc"]
        URL_GOLDENPATH = arguments["goldenpath"]
    else:
        GENOME = ""
        URL_UCSC = ""
        URL_GOLDENPATH = ""
    return SAMPLE, CONTROL, ALLELE, NAME, THREADS, GENOME, URL_UCSC, URL_GOLDENPATH


def _format_inputs(arguments: dict):
    SAMPLE, CONTROL, ALLELE, NAME, THREADS, GENOME, URL_UCSC, URL_GOLDENPATH = _parse_arguments(arguments)
    SAMPLE = preprocess.format_inputs.convert_to_posix_path(SAMPLE)
    CONTROL = preprocess.format_inputs.convert_to_posix_path(CONTROL)
    ALLELE = preprocess.format_inputs.convert_to_posix_path(ALLELE)

    SAMPLE_NAME: str = preprocess.format_inputs.extract_basename(SAMPLE)
    CONTROL_NAME: str = preprocess.format_inputs.extract_basename(CONTROL)
    FASTA_ALLELES: dict = preprocess.format_inputs.dictionize_allele(ALLELE)

    TEMPDIR = Path("DAJINResults", ".tempdir", NAME)
    SUBDIRS = ["cache", "fasta", "sam", "midsv", "clustering", "report", "result", "mutation_loci"]
    SUBDIRS_REPORT = ["HTML", "FASTA", "BAM", ".igvjs"]
    preprocess.format_inputs.make_directories(TEMPDIR, SUBDIRS, SUBDIRS_REPORT, SAMPLE_NAME, CONTROL_NAME)

    IS_CACHE_CONTROL = preprocess.validate_inputs.exists_cached_control(CONTROL, TEMPDIR)
    IS_CACHE_GENOME = preprocess.validate_inputs.exists_cached_genome(GENOME, TEMPDIR, IS_CACHE_CONTROL)
    if GENOME:
        if not IS_CACHE_GENOME:
            GENOME_COODINATES = preprocess.format_inputs.fetch_coodinate(GENOME, URL_UCSC, FASTA_ALLELES["control"])
            CHROME_SIZE = preprocess.format_inputs.fetch_chrom_size(GENOME_COODINATES["chr"], GENOME, URL_GOLDENPATH)
            preprocess.format_inputs.cache_coodinates_and_chromsize(TEMPDIR, GENOME, GENOME_COODINATES, CHROME_SIZE)
        else:
            GENOME_COODINATES = json.loads(Path(TEMPDIR, "cache", "genome_coodinates.jsonl").read_text())
            CHROME_SIZE = int(Path(TEMPDIR, "cache", "chrome_size.txt").read_text())
    return SAMPLE_NAME, CONTROL_NAME, FASTA_ALLELES, TEMPDIR, GENOME_COODINATES, CHROME_SIZE, THREADS


def _dtnow() -> str:
    return datetime.now().strftime("%Y-%m-%d %H:%M:%S")


def execute_control(arguments: dict):
    print(f"{_dtnow()}: {arguments['control']} is now processing...", file=sys.stderr)
    ###########################################################
    # Preprocess
    ###########################################################
    SAMPLE, CONTROL, ALLELE, NAME, THREADS, GENOME, _, _ = _parse_arguments(arguments)
    # preprocess.validate_inputs.check_files(SAMPLE, CONTROL, ALLELE)
    _, CONTROL_NAME, FASTA_ALLELES, TEMPDIR, GENOME_COODINATES, CHROME_SIZE, THREADS = _format_inputs(arguments)
    ###########################################################
    # Save Caches
    ###########################################################
    if Path(TEMPDIR, "report", "BAM", CONTROL_NAME, f"{CONTROL_NAME}.bam").exists():
        print(
            f"{arguments['control']} is already preprocessed and reuse the results for the current run...",
            file=sys.stderr,
        )
        return
    print(f"{_dtnow()}: Preprocess {arguments['control']}...", file=sys.stderr)
    # ============================================================
    # Export fasta files as single-FASTA format
    # ============================================================
    for identifier, sequence in FASTA_ALLELES.items():
        contents = "\n".join([">" + identifier, sequence]) + "\n"
        output_fasta = Path(TEMPDIR, "fasta", f"{identifier}.fasta")
        output_fasta.write_text(contents)
    print(f"{_dtnow()}: Mapping {arguments['control']}...", file=sys.stderr)
    # ============================================================
    # Mapping using mappy
    # ============================================================
    for path_fasta in Path(TEMPDIR, "fasta").glob("*.fasta"):
        name_fasta = path_fasta.stem
        preprocess.mappy_align.output_sam(TEMPDIR, path_fasta, name_fasta, CONTROL, CONTROL_NAME, threads=THREADS)
        preprocess.mappy_align.output_sam(
            TEMPDIR, path_fasta, name_fasta, CONTROL, CONTROL_NAME, preset="splice", threads=THREADS
        )
    # ============================================================
    # MIDSV conversion
    # ============================================================
    print(f"{_dtnow()}: Call MIDSV {arguments['control']}...", file=sys.stderr)
    preprocess.call_midsv(TEMPDIR, FASTA_ALLELES, CONTROL_NAME)
    ###########################################################
    # Save MIDSV and BAM
    ###########################################################
    # with open(Path(TEMPDIR, "midsv", f"{arguments['control']}.plk"), 'wb') as p:
    #     pickle.dump(midsv_control_alleles, p)
    report.report_bam.output_bam_control(TEMPDIR, CONTROL_NAME, GENOME, GENOME_COODINATES, CHROME_SIZE, THREADS)
    print(f"{_dtnow()}: \N{teacup without handle} {arguments['control']} is finished!", file=sys.stderr)


def execute_sample(arguments: dict):
    print(f"{_dtnow()}: {arguments['sample']} is now processing...", file=sys.stderr)
    ###########################################################
    # Preprocess
    ###########################################################
    SAMPLE, CONTROL, ALLELE, NAME, THREADS, GENOME, _, _ = _parse_arguments(arguments)
    # preprocess.validate_inputs.check_files(SAMPLE, CONTROL, ALLELE)
    SAMPLE_NAME, CONTROL_NAME, FASTA_ALLELES, TEMPDIR, GENOME_COODINATES, CHROME_SIZE, THREADS = _format_inputs(
        arguments
    )
    print(f"{_dtnow()}: Preprocess {arguments['sample']}...", file=sys.stderr)
    # ============================================================
    # Mapping with mappy
    # ============================================================
    for path_fasta in Path(TEMPDIR, "fasta").glob("*.fasta"):
        name_fasta = path_fasta.stem
        preprocess.mappy_align.output_sam(TEMPDIR, path_fasta, name_fasta, SAMPLE, SAMPLE_NAME, threads=THREADS)
        preprocess.mappy_align.output_sam(
            TEMPDIR, path_fasta, name_fasta, SAMPLE, SAMPLE_NAME, preset="splice", threads=THREADS
        )
    # ============================================================
    # MIDSV conversion
    # ============================================================
    preprocess.call_midsv(TEMPDIR, FASTA_ALLELES, SAMPLE_NAME)
    # ============================================================
    # Extract mutation loci
    # ============================================================
    MUTATION_LOCI_ALLELES = preprocess.extract_mutation_loci(TEMPDIR, FASTA_ALLELES, SAMPLE_NAME, CONTROL_NAME)
    with open(Path(TEMPDIR, "mutation_loci", f"{SAMPLE_NAME}.plk"), "wb") as p:
        pickle.dump(MUTATION_LOCI_ALLELES, p)
    KNOCKIN_LOCI_ALLELES = preprocess.extract_knockin_loci(TEMPDIR)
    # ============================================================
    # CSSPLITS Error Correction
    # ============================================================
    # preprocess.correct_knockin.execute(TEMPDIR, FASTA_ALLELES, CONTROL_NAME, SAMPLE_NAME)
    ########################################################################
    # Classify alleles
    ########################################################################
    print(f"{_dtnow()}: Classify {arguments['sample']}...", file=sys.stderr)
    classif_sample = classification.classify_alleles(TEMPDIR, FASTA_ALLELES, SAMPLE_NAME)
    ########################################################################
    # Clustering
    ########################################################################
    print(f"{_dtnow()}: Clustering {arguments['sample']}...", file=sys.stderr)
    clust_sample = clustering.add_labels(
        classif_sample, TEMPDIR, SAMPLE_NAME, CONTROL_NAME, MUTATION_LOCI_ALLELES, KNOCKIN_LOCI_ALLELES, THREADS
    )
    clust_sample = clustering.add_readnum(clust_sample)
    clust_sample = clustering.add_percent(clust_sample)
    clust_sample = clustering.update_labels(clust_sample)
    ########################################################################
    # Consensus call
    ########################################################################
    print(f"{_dtnow()}: Consensus calling {arguments['sample']}...", file=sys.stderr)
    # Downsampling to 1000 reads in each LABEL
    clust_subset_sample = consensus.subset_clust(clust_sample, 1000)
    cons_percentage, cons_sequence = consensus.call_consensus(clust_subset_sample, MUTATION_LOCI_ALLELES)
    allele_names = consensus.call_allele_name(cons_sequence, cons_percentage, FASTA_ALLELES)
    cons_percentage = consensus.update_key_by_allele_name(cons_percentage, allele_names)
    cons_sequence = consensus.update_key_by_allele_name(cons_sequence, allele_names)
    RESULT_SAMPLE = consensus.add_key_by_allele_name(clust_sample, allele_names)
    RESULT_SAMPLE.sort(key=lambda x: x["LABEL"])
    ########################################################################
    # Output Report：RESULT/FASTA/HTML/BAM/VCF
    ########################################################################
    # RESULT
    midsv.write_jsonl(RESULT_SAMPLE, Path(TEMPDIR, "result", f"{SAMPLE_NAME}.jsonl"))
    # FASTA
    for header, cons_seq in cons_sequence.items():
        cons_fasta = report.report_files.to_fasta(header, cons_seq)
        Path(TEMPDIR, "report", "FASTA", SAMPLE_NAME, f"{SAMPLE_NAME}_{header}.fasta").write_text(cons_fasta)
    # HTML
    for header, cons_per in cons_percentage.items():
        cons_html = report.report_files.to_html(header, cons_per)
        Path(TEMPDIR, "report", "HTML", SAMPLE_NAME, f"{SAMPLE_NAME}_{header}.html").write_text(cons_html)
    # BAM
    report.report_bam.output_bam_sample(
        TEMPDIR, RESULT_SAMPLE, SAMPLE_NAME, GENOME, GENOME_COODINATES, CHROME_SIZE, THREADS
    )
    for path_bam_igvjs in Path(TEMPDIR, "cache", ".igvjs").glob(f"{CONTROL_NAME}_control.bam*"):
        shutil.copy(path_bam_igvjs, Path(TEMPDIR, "report", ".igvjs", SAMPLE_NAME))
    # VCF
    # working in progress
    print(f"{_dtnow()}: \N{teacup without handle} {arguments['sample']} is finished!", file=sys.stderr)



from __future__ import annotations

import re
from collections import defaultdict
from itertools import groupby


def _call_percentage(cssplits: list[str], mutation_loci) -> list[dict[str, float]]:
    """call position weight matrix in defferent loci.
    - non defferent loci are annotated to "Match" or "Unknown(N)"
    - sequence errors are annotated to "SEQERROR"
    """
    cssplits_transposed = [list(cs) for cs in zip(*cssplits)]
    coverage = len(cssplits)
    cons_percentage = []
    for cs_transposed, mut_loci in zip(cssplits_transposed, mutation_loci):
        count_cs = defaultdict(float)
        for cs in cs_transposed:
            if cs[0] in {"+", "-", "*"} and cs[0] not in mut_loci:
                cs = "SEQERROR"
            count_cs[cs] += 1 / coverage * 100
        count_cs_sorted = dict(sorted(count_cs.items(), key=lambda x: x[1], reverse=True))
        cons_percentage.append(count_cs_sorted)
    return cons_percentage


def _replace_percentage(cons_percentage: list[dict[str, float]]) -> list[dict[str, float]]:
    """replace sequence error as distributing according to proportion of cs tags"""
    cons_percentage_updated = []
    for dict_percentage in cons_percentage:
        if "SEQERROR" not in dict_percentage:
            cons_percentage_updated.append(dict_percentage)
            continue
        if len(dict_percentage) == 1 and dict_percentage["SEQERROR"]:
            cons_percentage_updated.append({"N": 100})
            continue
        dict_percentage_update = dict()
        div = 100 / (100 - dict_percentage["SEQERROR"])
        for key, val in dict_percentage.items():
            if key == "SEQERROR":
                continue
            dict_percentage_update[key] = round(val * div, 5)
        cons_percentage_updated.append(dict_percentage_update)
    return cons_percentage_updated


def _call_sequence(cons_percentage: list[dict[str, float]]) -> str:
    consensus_sequence = []
    for cons_per in cons_percentage:
        cons = max(cons_per, key=cons_per.get)
        if cons.startswith("="):
            cons = cons.replace("=", "")
        elif cons.startswith("-"):
            continue
        elif cons.startswith("*"):
            cons = cons[-1]
        elif cons.startswith("+"):
            cons_ins = cons.split("|")
            if cons_ins[-1].startswith("="):
                cons = cons.replace("=", "")
            elif cons_ins[-1].startswith("-"):
                cons = "".join(cons_ins[:-1])
            elif cons_ins[-1].startswith("*"):
                cons = "".join([*cons_ins[:-1], cons_ins[-1][-1]])
            cons = cons.replace("+", "")
            cons = cons.replace("|", "")
        consensus_sequence.append(cons)
    return "".join(consensus_sequence)


def _detect_sv(cons_percentage: defaultdict[list], threshold: int = 50) -> list[bool]:
    exists_sv = []
    for cons_per in cons_percentage.values():
        cons_cssplits = []
        for cssplit in cons_per:
            seq = max(cssplit, key=cssplit.get)
            cons_cssplits.append(seq)
        cons_cssplits = "".join(cons_cssplits)
        if "N" * threshold in cons_cssplits:
            exists_sv.append(True)
        elif re.search(rf"(\+[ACGTN]\|){{{threshold}}}", cons_cssplits):
            exists_sv.append(True)
        elif re.search(rf"(\-[ACGTN]){{{threshold}}}", cons_cssplits):
            exists_sv.append(True)
        elif re.search(rf"(\*[ACGTN][ACGTN]){{{threshold}}}", cons_cssplits):
            exists_sv.append(True)
        elif re.search(r"[acgtn]", cons_cssplits):
            exists_sv.append(True)
        else:
            exists_sv.append(False)
    return exists_sv


###########################################################
# main
###########################################################


def call_consensus(clust_sample: list[dict], MUTATION_LOCI_ALLELES) -> tuple[defaultdict[list], defaultdict[str]]:
    cons_percentage = defaultdict(list)
    cons_sequence = defaultdict(str)
    clust_sample.sort(key=lambda x: x["LABEL"])
    for _, group in groupby(clust_sample, key=lambda x: x["LABEL"]):
        clust = list(group)
        keys = (
            clust[0]["ALLELE"],
            clust[0]["LABEL"],
            clust[0]["PERCENT"],
        )
        mutation_loci = MUTATION_LOCI_ALLELES[clust[0]["ALLELE"]]
        cssplits = [cs["CSSPLIT"].split(",") for cs in clust]
        cons_per = _call_percentage(cssplits, mutation_loci)
        cons_per = _replace_percentage(cons_per)
        cons_seq = _call_sequence(cons_per)
        cons_percentage[keys] = cons_per
        cons_sequence[keys] = cons_seq
    return cons_percentage, cons_sequence


def call_allele_name(
    cons_sequence: defaultdict[dict], cons_percentage: defaultdict[list], FASTA_ALLELES: dict
) -> dict[int, str]:
    exists_sv = _detect_sv(cons_percentage)
    label_digits = len(str(len(cons_percentage)))
    allele_names = {}
    for is_sv, (keys, cons_seq) in zip(exists_sv, cons_sequence.items()):
        ALLELE, LABEL, PERCENT = keys
        label_format = f"{LABEL:0{label_digits}}"
        allele_name = f"allele{label_format}_{ALLELE}"
        if cons_seq == FASTA_ALLELES[ALLELE]:
            allele_name += "_intact"
        elif is_sv:
            allele_name += "_sv"
        else:
            allele_name += "_indels"
        allele_name += f"_{PERCENT}%"
        allele_names.update({LABEL: allele_name})
    return allele_names


def update_key_by_allele_name(cons: dict, allele_names: dict[int, str]) -> dict:
    for key, allele_name in zip(list(cons.keys()), allele_names.values()):
        cons[allele_name] = cons.pop(key)
    return cons


def add_key_by_allele_name(clust_sample: list[dict], allele_names: dict[int, str]) -> list[dict]:
    for clust in clust_sample:
        label = clust["LABEL"]
        clust["NAME"] = allele_names[label]
    return clust_sample



from __future__ import annotations

import sys
from itertools import groupby
from pathlib import Path

import pandas as pd
import wslPath

from DAJIN2.core import core_execute
from DAJIN2.postprocess import report
from DAJIN2.preprocess.validate_inputs import (validate_files,
                                               validate_genome_and_fetch_urls)


def execute_batch_mode(arguments: dict[str]):
    threads = int(arguments["threads"])
    path_batchfile = arguments["file"]

    ###############################################################################
    # Validate batch file
    ###############################################################################
    try:
        path_batchfile = wslPath.toPosix(path_batchfile)
    except ValueError:
        pass

    # ----------------------------------------------------------
    # Check file exists
    # ----------------------------------------------------------
    if not Path(path_batchfile).exists():
        raise FileNotFoundError(f"'{path_batchfile}' does not exist.")

    # ----------------------------------------------------------
    # Load the file
    # ----------------------------------------------------------
    try:
        df_batchfile = pd.read_excel(path_batchfile)
        inputs = []
        inputs.append(df_batchfile.columns.to_list())
        inputs += df_batchfile.values.tolist()
    except ValueError:
        inputs = [s.split(",") for s in Path(path_batchfile).read_text().strip().split("\n")]

    ################################################################################
    # Validate Column of the batch file
    ################################################################################
    columns = inputs[0]
    columns_set = set(columns)
    columns_required = set(["sample", "control", "allele", "name"])
    columns_all = set(["sample", "control", "allele", "name", "genome"])

    if not columns_required.issubset(columns_set):
        raise ValueError(f"{path_batchfile} must contain 'sample', 'control', and 'allele' in the header")

    if not columns_set.issubset(columns_all):
        raise ValueError(
            f"Accepted header names of {path_batchfile} are 'sample', 'control', 'allele', 'name', and 'genome'."
        )

    ################################################################################
    # Validate contents in the batch file
    ################################################################################
    index_name = columns.index("name")
    contents = inputs[1:]
    contents.sort(key=lambda x: x[index_name])

    for name, groups in groupby(contents, key=lambda x: x[index_name]):
        for group in groups:
            args = {h: g for h, g in zip(columns, group)}
            validate_files(args["sample"], args["control"], args["allele"])
            if "genome" in args:
                UCSC_URL, GOLDENPATH_URL = validate_genome_and_fetch_urls(args["genome"])
            else:
                UCSC_URL, GOLDENPATH_URL = None, None
    ##############################################################################
    # Perform DAJIN
    ##############################################################################

    for name, groups in groupby(contents, key=lambda x: x[index_name]):
        done_controle = False
        for group in groups:
            args = {h: g for h, g in zip(columns, group)}
            args["threads"] = threads
            if done_controle == False:
                print(f"{args['control']} is now processing...", file=sys.stderr)
                core_execute.execute_control(args)
                done_controle = True
            print(f"{args['sample']} is now processing...", file=sys.stderr)
            core_execute.execute_sample(args)
        report.report(name)
        print(f"Finished! Open DAJINResults/{name} to see the report.", file=sys.stderr)



from __future__ import annotations

from collections import defaultdict
from itertools import groupby, permutations
from pathlib import Path

import midsv
import json
from typing import Generator
from DAJIN2.core.preprocess import mappy_align


def _extract_diff_loci(TEMPDIR: Path) -> defaultdict[dict]:
    """
    Extract differencial loci between alleles
        - The purpose is to lower match_score between very similar alleles such as point mutation.
    """
    fasta_alleles = list(Path(TEMPDIR, "fasta").iterdir())
    fasta_alleles = [f for f in fasta_alleles if f.suffix != ".fai"]
    mutation_alleles = defaultdict(dict)
    for pair in list(permutations(fasta_alleles, 2)):
        ref, query = pair
        ref_allele = ref.stem
        alignments = mappy_align.to_sam(ref, query, preset="splice")
        alignments = list(alignments)
        alignments = [a.split("\t") for a in alignments]
        alignments_midsv = midsv.transform(alignments, midsv=False, cssplit=True, qscore=False)[0]
        cssplits = alignments_midsv["CSSPLIT"].split(",")
        mutations = dict()
        for i, cs in enumerate(cssplits):
            if cs.startswith("="):
                continue
            mutations.update({i: cs})
        if len(mutations) < 10:
            mutation_alleles[ref_allele].update(mutations)
    return mutation_alleles


def _calc_match(CSSPLIT: str, mutations: dict) -> float:
    match_score = CSSPLIT.count("=")
    match_score -= CSSPLIT.count("+")  # insertion
    match_score -= sum(cs.islower() for cs in CSSPLIT)  # inversion
    cssplit = CSSPLIT.split(",")
    for i, mut in mutations.items():
        if cssplit[i] == mut:
            match_score = 0
    return match_score / len(cssplit)


def read_json(filepath) -> Generator[dict[str, str]]:
    with open(filepath, "r") as f:
        for line in f:
            yield json.loads(line)


###########################################################
# main
###########################################################


def classify_alleles(TEMPDIR: Path, FASTA_ALLELES: dict, SAMPLE_NAME: str) -> list[dict]:
    mutations = _extract_diff_loci(TEMPDIR)
    # Scoring
    score_of_each_alleles = []
    for allele in FASTA_ALLELES:
        midsv_sample = read_json(Path(TEMPDIR, "midsv", f"{SAMPLE_NAME}_{allele}.json"))
        for dict_midsv in midsv_sample:
            score = _calc_match(dict_midsv["CSSPLIT"], mutations[allele])
            dict_midsv.update({"SCORE": score})
            dict_midsv.update({"ALLELE": allele})
            score_of_each_alleles.append(dict_midsv)
    score_of_each_alleles.sort(key=lambda x: x["QNAME"])
    # Extract alleles with max scores
    possible_allele = []
    for _, group in groupby(score_of_each_alleles, key=lambda x: x["QNAME"]):
        max_score = -float("inf")
        for readinfo in group:
            if readinfo["SCORE"] > max_score:
                max_score = readinfo["SCORE"]
                max_read = readinfo
                del max_read["SCORE"]
        possible_allele.append(max_read)
    return possible_allele



from DAJIN2.core.consensus.consensus import call_consensus
from DAJIN2.core.consensus.consensus import call_allele_name
from DAJIN2.core.consensus.consensus import update_key_by_allele_name
from DAJIN2.core.consensus.consensus import add_key_by_allele_name
from DAJIN2.core.consensus.subset import subset_clust



from __future__ import annotations

import os
import argparse
import multiprocessing
import sys
from itertools import groupby, islice
from pathlib import Path

import pandas as pd
import wslPath

from DAJIN2 import gui, view
from DAJIN2.core import core_execute
from DAJIN2.postprocess import report
from DAJIN2.preprocess.validate_inputs import validate_files, validate_genome_and_fetch_urls

VERSION = "0.2.0"

# prevent BLAS from using all cores
os.environ["OMP_NUM_THREADS"] = "1"
os.environ["OPENBLAS_NUM_THREADS"] = "1"
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["VECLIB_MAXIMUM_THREADS"] = "1"
os.environ["NUMEXPR_NUM_THREADS"] = "1"


def _execute_single_mode(arguments: dict[str]):
    ################################################################################
    # Validate contents
    ################################################################################
    validate_files(arguments["sample"], arguments["control"], arguments["allele"])
    URLS_GENOME: dict[str, dict[str, str]] = dict()
    if "genome" in arguments:
        URLS_GENOME.update({arguments["genome"]: validate_genome_and_fetch_urls(arguments["genome"])})
        arguments.update(URLS_GENOME[arguments["genome"]])
    ##############################################################################
    # Perform DAJIN2
    ##############################################################################
    core_execute.execute_control(arguments)
    core_execute.execute_sample(arguments)
    name = arguments["name"]
    report.report(name)
    print(
        f"\N{party popper} Finished! Open DAJINResults/{name} to see the report.",
        file=sys.stderr,
    )


def _update_threads(threads) -> int:
    threads_updated = min(int(threads), os.cpu_count() - 1)
    threads_updated = max(1, threads_updated)
    return threads_updated


def _extract_unique_contents(list: list) -> list:
    list_unique = []
    for content in list:
        if content not in list_unique:
            list_unique.append(content)
    return list_unique


def _batched(iterable, chunk_size):
    iterator = iter(iterable)
    chunk = tuple(islice(iterator, chunk_size))
    while chunk:
        yield chunk
        chunk = tuple(islice(iterator, chunk_size))


def _run_multiprocess(function, arguments: list, num_workers: int = 1) -> None:
    arguments_batched = _batched(arguments, num_workers)
    for args in arguments_batched:
        jobs = []
        for arg in args:
            p = multiprocessing.Process(target=function, args=(arg,))
            jobs.append(p)
            p.start()
        for job in jobs:
            if job.exitcode == 1:
                sys.exit(1)
            job.join()
    return


def _execute_batch_mode(arguments: dict[str]):
    path_batchfile = arguments["file"]
    ###############################################################################
    # Validate batch file
    ###############################################################################
    try:
        path_batchfile = wslPath.toPosix(path_batchfile)
    except ValueError:
        pass
    # ----------------------------------------------------------
    # Check file exists
    # ----------------------------------------------------------
    if not Path(path_batchfile).exists():
        raise FileNotFoundError(f"'{path_batchfile}' does not exist.")
    # ----------------------------------------------------------
    # Load the file
    # ----------------------------------------------------------
    try:
        df_batchfile = pd.read_excel(path_batchfile)
        inputs = []
        inputs.append(df_batchfile.columns.to_list())
        inputs += df_batchfile.values.tolist()
    except ValueError:
        inputs = [s.split(",") for s in Path(path_batchfile).read_text().strip().split("\n")]
    ################################################################################
    # Validate Column of the batch file
    ################################################################################
    columns = inputs[0]
    if not {"sample", "control", "allele", "name"}.issubset(set(columns)):
        raise ValueError(f"{path_batchfile} must contain 'sample', 'control', and 'allele' in the header")
    if not set(columns).issubset({"sample", "control", "allele", "name", "genome"}):
        raise ValueError(
            f"Accepted header names of {path_batchfile} are 'sample', 'control', 'allele', 'name', and 'genome'."
        )
    ################################################################################
    # Validate contents in the batch file
    ################################################################################
    index_name = columns.index("name")
    contents = inputs[1:]
    contents.sort(key=lambda x: x[index_name])
    for _, groups in groupby(contents, key=lambda x: x[index_name]):
        for group in groups:
            args = {h: g for h, g in zip(columns, group)}
            validate_files(args["sample"], args["control"], args["allele"])
            URLS_GENOME: dict[str, dict[str, str]] = dict()
            if "genome" in args and not args["genome"] in URLS_GENOME:
                URLS_GENOME.update({args["genome"]: validate_genome_and_fetch_urls(args["genome"])})
    ##############################################################################
    # Perform DAJIN2
    ##############################################################################
    num_workers = _update_threads(arguments["threads"])
    index_name = columns.index("name")
    contents.sort(key=lambda x: x[index_name])
    contents_sample = []
    for name, groups in groupby(contents, key=lambda x: x[index_name]):
        groups = list(groups)
        # ------------------------------
        # Handle controls
        # ------------------------------
        contents_control = []
        for group in groups:
            args = {h: g for h, g in zip(columns, group)}
            args["sample"] = args["control"]
            if len(groups) > 1:
                args["threads"] = 1
            if "genome" in args:
                args.update(URLS_GENOME[args["genome"]])
            contents_control.append(args)
        contents_control_unique = _extract_unique_contents(contents_control)
        _run_multiprocess(core_execute.execute_control, contents_control_unique, num_workers)
        # ------------------------------
        # Handle samples
        # ------------------------------
        contents_sample = []
        for group in groups:
            args = {h: g for h, g in zip(columns, group)}
            if args["sample"] == args["control"]:
                continue
            if "threads" not in args:
                args["threads"] = 1
            args.update(URLS_GENOME[args["genome"]])
            contents_sample.append(args)
        contents_sample_unique = _extract_unique_contents(contents_sample)
        _run_multiprocess(core_execute.execute_sample, contents_sample_unique, num_workers)
        report.report(name)
        print(
            f"\N{party popper} Finished! Open DAJINResults/{name} to see the report.",
            file=sys.stderr,
        )


def main():
    parser = argparse.ArgumentParser()

    ###############################################################################
    # Single mode
    ###############################################################################

    parser.add_argument("-s", "--sample", type=str, help="Full path to a sample FASTQ file")
    parser.add_argument("-c", "--control", type=str, help="Full path to a control FASTQ file")
    parser.add_argument("-a", "--allele", type=str, help="Full path to a FASTA file")
    parser.add_argument("-n", "--name", type=str, help="Output directory name")
    parser.add_argument(
        "-g", "--genome", type=str, default="", help="Reference genome ID (e.g hg38, mm10) [default: '']"
    )
    parser.add_argument("-t", "--threads", type=int, default=1, help="Number of threads [default: 1]")
    parser.add_argument("-v", "--version", action="version", version=f"DAJIN2 version {VERSION}")
    parser.add_argument("-d", "--debug", action="store_true", help=argparse.SUPPRESS)

    ###############################################################################
    # Batch mode
    ###############################################################################

    def batchmode(args):
        arguments = dict()
        arguments["file"] = args.file
        arguments["threads"] = int(args.threads)
        arguments["debug"] = args.debug
        _execute_batch_mode(arguments)

    subparser = parser.add_subparsers()
    parser_batch = subparser.add_parser("batch", help="DAIJN2 batch mode")
    parser_batch.add_argument("-f", "--file", required=True, type=str, help="CSV or Excel file.")
    parser_batch.add_argument("-t", "--threads", default=1, type=int, help="Number of threads [default: 1]")
    parser_batch.add_argument("-d", "--debug", action="store_true", help=argparse.SUPPRESS)
    parser_batch.set_defaults(handler=batchmode)

    ###############################################################################
    # GUI mode
    ###############################################################################

    def guimode(args):
        gui.execute()

    parser_gui = subparser.add_parser("gui", help="DAIJN2 GUI mode")
    parser_gui.set_defaults(handler=guimode)

    ###############################################################################
    # View mode
    ###############################################################################

    def viewmode(args):
        view.execute(args.name)

    parser_view = subparser.add_parser("view", help="DAIJN2 View mode to launch igvjs")
    parser_view.add_argument("-n", "--name", required=True, type=str, help="Output name of the report")
    parser_view.set_defaults(handler=viewmode)

    ###############################################################################
    # Parse arguments
    ###############################################################################

    args = parser.parse_args()

    if hasattr(args, "handler"):
        args.handler(args)
    else:
        if args.sample is None:
            raise AttributeError("the following arguments are required: -s/--sample")
        if args.control is None:
            raise AttributeError("the following arguments are required: -c/--control")
        if args.allele is None:
            raise AttributeError("the following arguments are required: -a/--allele")
        if args.name is None:
            raise AttributeError("the following arguments are required: -n/--name")
        arguments = dict()
        arguments["sample"] = args.sample
        arguments["control"] = args.control
        arguments["allele"] = args.allele
        arguments["name"] = args.name
        if args.genome:
            arguments["genome"] = args.genome
        arguments["threads"] = int(args.threads)
        arguments["debug"] = args.debug
        _execute_single_mode(arguments)


if __name__ == "__main__":
    main()



from __future__ import annotations

from collections import defaultdict
from itertools import groupby
from typing import Generator
import json
from pathlib import Path
from DAJIN2.core.clustering.make_score import make_score
from DAJIN2.core.clustering.return_labels import return_labels
import random

# def _compress_insertion(cssplits: Generator[list[str]]) -> Generator[dict[str, int]]:
#     """Insertion will be subdivided by sequence error in the its sequence,
#     so it is compressed as a '+I' to eliminate mutations.
#     #TODO ただ、これでは、insertion配列の中に真のmutationがある場合に
#     #TODO そのmutationを抽出できないので**insertion配列の中にmutationがある場合は
#     #TODO insertion配列をそのまま残す**必要がある。
#     """
#     for cssplit in cssplits:
#         for i, cs in enumerate(cssplit):
#             if cs.startswith("+"):
#                 cssplit[i] = "+I" + cs.split("|")[-1]
#         yield cssplit


def _generate_mutation_kmers(midsv_sample: Generator[list[str]], mutation_loci: list[set[str]]) -> Generator[list[str]]:
    for cssplit in (cs["CSSPLIT"].split(",") for cs in midsv_sample):
        cs_mutation = ["N,N,N"]
        for i in range(1, len(cssplit) - 1):
            if mutation_loci[i] == set():
                cs_mutation.append("N,N,N")
                continue
            mutation = cssplit[i][0]  # +, - , *, =, N
            """Insertion will be subdivided by sequence error in the its sequence,
            so it is compressed as a '+I' to eliminate mutations.
            #TODO ただ、これでは、insertion配列の中に真のmutationがある場合に
            #TODO そのmutationを抽出できないので**insertion配列の中にmutationがある場合は
            #TODO insertion配列をそのまま残す**必要がある。
            """
            if mutation == "+":
                cssplit[i] = "+I" + cssplit[i].split("|")[-1]
            if mutation in mutation_loci[i]:
                kmer = ",".join([cssplit[i - 1], cssplit[i], cssplit[i + 1]])
                cs_mutation.append(kmer)
            else:
                cs_mutation.append("N,N,N")
        cs_mutation.append("N,N,N")
        yield cs_mutation


def _annotate_score(cssplits: Generator[list[str]], mutation_score: list[dict[str:float]]) -> Generator[list[float]]:
    for cssplit in cssplits:
        score = [0 for _ in range(len(cssplit))]
        for i, (cs, mutscore) in enumerate(zip(cssplit, mutation_score)):
            if mutscore == {}:
                continue
            if cs in mutscore:
                score[i] = mutscore[cs]
        yield score


def _reorder_labels(labels: list[int], start: int = 0) -> list[int]:
    labels_ordered = labels.copy()
    num = start
    d = defaultdict(int)
    for i, l in enumerate(labels_ordered):
        if not d[l]:
            num += 1
            d[l] = num
        labels_ordered[i] = d[l]
    return labels_ordered


def read_json(filepath: Path | str) -> Generator[dict[str, str]]:
    with open(filepath, "r") as f:
        for line in f:
            yield json.loads(line)


def write_json(filepath: Path | str, data: Generator) -> None:
    with open(filepath, "w") as f:
        for line in data:
            f.write(json.dumps(line) + "\n")


def add_labels(
    classif_sample, TEMPDIR, SAMPLE_NAME, CONTROL_NAME, MUTATION_LOCI_ALLELES, KNOCKIN_LOCI_ALLELES, THREADS: int = 1
) -> list[dict[str]]:
    labels_all = []
    max_label = 0
    classif_sample.sort(key=lambda x: x["ALLELE"])
    for allele, group in groupby(classif_sample, key=lambda x: x["ALLELE"]):
        RANDOM_NUM = random.randint(0, 10**10)
        mutation_loci: dict[str, set[str]] = MUTATION_LOCI_ALLELES[allele]
        if all(m == set() for m in mutation_loci):
            labels = [1] * len(classif_sample)
            labels_reorder = _reorder_labels(labels, start=max_label)
            max_label = max(labels_reorder)
            labels_all.extend(labels_reorder)
            continue
        if allele in KNOCKIN_LOCI_ALLELES:
            knockin_loci = KNOCKIN_LOCI_ALLELES[allele]
        else:
            knockin_loci = set()
        path_sample = Path(TEMPDIR, "clustering", f"{SAMPLE_NAME}_{allele}_{RANDOM_NUM}.json")
        path_control = Path(TEMPDIR, "midsv", f"{CONTROL_NAME}_{allele}.json")
        write_json(path_sample, group)
        mutation_score = make_score(
            _generate_mutation_kmers(read_json(path_sample), mutation_loci),
            _generate_mutation_kmers(read_json(path_control), mutation_loci),
            mutation_loci,
            knockin_loci,
        )
        scores_sample = _annotate_score(_generate_mutation_kmers(read_json(path_sample), mutation_loci), mutation_score)
        scores_control = _annotate_score(
            _generate_mutation_kmers(read_json(path_control), mutation_loci), mutation_score
        )
        path_score_sample = Path(TEMPDIR, "clustering", f"{SAMPLE_NAME}_{allele}_score_{RANDOM_NUM}.json")
        path_score_control = Path(TEMPDIR, "clustering", f"{CONTROL_NAME}_{allele}_score_{RANDOM_NUM}.json")
        write_json(path_score_sample, scores_sample)
        write_json(path_score_control, scores_control)
        labels = return_labels(path_score_sample, path_score_control)
        labels_reorder = _reorder_labels(labels, start=max_label)
        max_label = max(labels_reorder)
        labels_all.extend(labels_reorder)
        # Remove temporary files
        path_sample.unlink()
        path_score_sample.unlink()
        path_score_control.unlink()
    clust_sample = classif_sample.copy()
    for clust, label in zip(clust_sample, labels_all):
        clust["LABEL"] = label
    return clust_sample


def add_readnum(clust_sample: list[dict]) -> list[dict]:
    clust_result = clust_sample.copy()
    readnum = defaultdict(int)
    for cs in clust_result:
        readnum[cs["LABEL"]] += 1
    for cs in clust_result:
        cs["READNUM"] = readnum[cs["LABEL"]]
    return clust_result


def add_percent(clust_sample: list[dict]) -> list[dict]:
    clust_result = clust_sample.copy()
    n_sample = len(clust_result)
    percent = defaultdict(int)
    for cs in clust_result:
        percent[cs["LABEL"]] += 1 / n_sample
    percent = {key: round(val * 100, 3) for key, val in percent.items()}
    for cs in clust_result:
        cs["PERCENT"] = percent[cs["LABEL"]]
    return clust_result


def update_labels(clust_sample: list[dict]) -> list[dict]:
    """
    Allocate new labels according to the ranking by PERCENT
    """
    clust_result = clust_sample.copy()
    clust_result.sort(key=lambda x: (-x["PERCENT"], x["LABEL"]))
    new_label = 1
    prev_label = clust_result[0]["LABEL"]
    for cs in clust_result:
        if prev_label != cs["LABEL"]:
            new_label += 1
        prev_label = cs["LABEL"]
        cs["LABEL"] = new_label
    return clust_result
